"""Automatically generated by hassfest.

To update, run python3 -m script.hassfest
"""


ZEROCONF = {
    "_axis-video._tcp.local.": [
        "axis"
    ],
    "_coap._udp.local.": [
        "tradfri"
    ],
    "_esphomelib._tcp.local.": [
        "esphome"
    ],
    "_googlecast._tcp.local.": [
        "cast"
    ],
    "_hap._tcp.local.": [
        "homekit_controller"
    ]
}

HOMEKIT = {
    "BSB002": "hue",
    "LIFX": "lifx",
    "TRADFRI": "tradfri",
    "Wemo": "wemo"
}
